/* $Id: table.c,v 1.1 1996/10/24 04:27:40 ryo freeze $
 *
 *	\[XR[hWFl[^
 *	e[uW[
 *	Copyright (C) 1989,1990 K.Abe
 *	All rights reserved.
 *	Copyright (C) 1997,1998 Tachibana Eriko.
 *
 */

#include <stdio.h>
#include <stdlib.h>	/* atoi */
#include <string.h>
#ifdef unix
    #include <ctype.h>
    #include <unistd.h>
#else
    #ifdef MYU_HEADERS
	#include <ctype.h>	/* tolower */
    #else
	#include <fctype.h>	/* ctype is bad because macro expansion */
    #endif			/* causes twice evaluation. */
#endif

#include "estruct.h"
#include "etc.h"
#include "global.h"
#include "hex.h"		/* strend */
#include "label.h"
#include "labelfile.h"
#include "offset.h"		/* depend_address */
#include "table.h"


USEOPTION	option_q;

extern int	Quiet_mode;


static table*	Table;
static int	TableCounter;
static int	Line_num;


/* table.loop mode */
#define TIMES_AUTOMATIC		0
#define TIMES_DECIDE_BY_BREAK	-1
#define eprintfq if( option_q && Quiet_mode < 1 ) eprintf


#ifdef DEBUG
#define debug(str)	printf(str)
#else
#define debug(str)
#endif


private void	interpret( table *table_ptr )
{
    extern int	yyparse (void);

    formula*	cur_expr = table_ptr->formulaptr;
    address	pc0 = 0;
    int		i, loop = 0;
    lblbuf*	lblptr = NULL;
    boolean	in_bss;
    address	limit;

    Eval_TableTop = Eval_PC = table_ptr->tabletop;
    ParseMode = PARSE_ANALYZING;

    in_bss = (BeginBSS <= Eval_PC) ? TRUE : FALSE;
    limit = in_bss ? Last : BeginBSS;

#if 0
    if (BeginBSS <= Eval_PC)
#else
    if (Last <= Eval_PC)	/* BSS ɂe[uF߂ */
#endif
	return;

    do {
	for( i = 0 ; i < table_ptr->lines ; i++ ) {
	    Eval_Count = 0;		/* First, Eval_Count is set in yyparse() */
	    do {
		Lexptr = cur_expr[ i ].expr;
		eprintfq( "[%d:%06x]:%s\n" , loop + 1 , Eval_PC , Lexptr );
		if( yyparse() == 1 )
		    err ("Syntax error (%d s)\n", cur_expr[ i ].line);

		cur_expr[ i ].id = Eval_ID;
		pc0 = Eval_PC;

		if( depend_address( Eval_PC ) &&
		    Eval_ID != LONGSIZE && Eval_ID != EVENID &&
#if 0
		    Eval_ID != LONGID &&
#endif
		    Eval_ID != BREAKID && Eval_ID != CRID ) {
		    if( table_ptr->loop == TIMES_AUTOMATIC ) {
			debug( "depend chk" );
			goto tableend;
		    }
		    err ("AhXˑŃO[hw肳Ă܂(%x)\n", Eval_PC);
		}

#define BSS_CHECK(idstr) ({ if (in_bss) goto bss_error; })
#define ODD_CHECK(size) ({ \
	if ((int)Eval_PC & 1) \
	    eprintf ("Warning: AhX(%x)%sw肳Ă܂.\n", Eval_PC, size); \
	})

		switch (Eval_ID) {
		case LONGSIZE:
		    ODD_CHECK ("O[h");
		    Eval_PC += 4;
		    break;
		case WORDSIZE:
		    ODD_CHECK ("[h");
		    Eval_PC += 2;
		    break;
		case BYTESIZE:
		    Eval_PC++;
		    break;

		case SINGLESIZE:
		    ODD_CHECK ("Px");
		    Eval_PC += 4;
		    break;
		case DOUBLESIZE:
		    ODD_CHECK ("{x");
		    Eval_PC += 8;
		    break;
		case EXTENDSIZE:
		    ODD_CHECK ("gx");
		    Eval_PC += 12;
		    break;
		case PACKEDSIZE:
		    ODD_CHECK ("pbNhfV}");
		    Eval_PC += 12;
		    break;
#if 0
		case LONGID:
		    BSS_CHECK ("long");
		    ODD_CHECK ("O[h");
		    Eval_PC += Eval_Bytes*4;
		    break;
		case WORDID:
		    BSS_CHECK ("word");
		    ODD_CHECK ("[h");
		    Eval_PC += Eval_Bytes*2;
		    break;
#endif
		case BYTEID:
		    BSS_CHECK ("byte");
		    Eval_PC += Eval_Bytes;
		    break;
		case ASCIIID:
		    BSS_CHECK ("ascii");
		    Eval_PC += Eval_Bytes;
		    break;
		case ASCIIZID:
		    BSS_CHECK ("asciiz");
		    while (PEEK_BYTE (Eval_PC))
			Eval_PC++;
		    break;
		case LASCIIID:
		    BSS_CHECK ("lascii");
		    Eval_PC += PEEK_BYTE (Eval_PC) + 1;
		    break;
		case EVENID:
		    Eval_PC += ( (long)Eval_PC & 1 );
		    break;
		case CRID:
		    break;
		case BREAKID:
		    if( Eval_Break ) {
			debug( "break" );
			loop++;
			goto tableend;
		    }
		    break;
		default:	/* reduce warning message */
		    break;
		}
#undef	ODD_CHECK
#undef	BSS_CHECK

		if( limit <= Eval_PC ||
		    ( table_ptr->loop == TIMES_AUTOMATIC &&
		     ( ( lblptr = next( pc0 + 1 ) ) == NULL ||
		      Eval_PC >= lblptr->label ) ) ) {
		    loop++;
		    /*  _________ łH */
		    if (lblptr && Eval_PC > lblptr->label)
			Eval_PC = pc0;
		    debug( "exceeded next label" );
		    goto tableend;
		}
#if 0
		if (Eval_ID != BREAKID && Eval_ID != CRID)
		    if (!regist_label (pc0, DATLABEL | Eval_ID))
			   eprintf ("??? address %x", pc0);
#endif
		Eval_Count--;
	    } while( Eval_Count > 0 );
	}
	loop++;
    } while ((Eval_PC < limit &&
	      (table_ptr->loop == TIMES_AUTOMATIC &&
	       (lblptr = next (pc0 + 1)) != NULL && Eval_PC < lblptr->label)
	     )
	     || table_ptr->loop == TIMES_DECIDE_BY_BREAK
	     || (table_ptr->loop != TIMES_AUTOMATIC && loop < table_ptr->loop));

tableend:
    eprintfq( "Table  %x - %x (%d)Ɣf܂\n" ,
	    Eval_TableTop , Eval_PC , loop );
    if( !regist_label( Eval_PC , DATLABEL|ENDTABLE|UNKNOWN ) ) {
	/* printf( "??? address %x\n" , Eval_PC ); */
	;
    }
    if( table_ptr->loop == TIMES_AUTOMATIC || table_ptr->loop == TIMES_DECIDE_BY_BREAK )
	table_ptr->loop = loop;			/* set implicit loop times */
    return;

bss_error:
    err ("BSS ł͎gpłȂʎqł(%d s)\n", cur_expr[i].line);
}


private void	interpret_all( void )
{
    int		i;

    for( i = 0 ; i < TableCounter ; i++ )
	interpret( &Table[ i ] );
}


/*

  e[ȕ

*/
private void	tablejob( address tabletop , FILE *fp )
{
    formula*	cur_expr;
    int		formula_line = 0;

    regist_label( tabletop , DATLABEL|TABLE|FORCE );

#ifdef	STATIC_TABLE_BUF
    if (TableCounter == MAX_TABLE_NUM)
	err ("e[u %d z܂.\n", MAX_TABLE_NUM);
#else
    /* e[uobt@m */
    Table = Realloc (Table, sizeof (table) * (TableCounter + 1));
#endif

    Table[ TableCounter ].tabletop = tabletop;

#ifdef	STATIC_TABLE_BUF
    if ((cur_expr = malloc (sizeof (formula) * MAX_TABLE_LINE)) == NULL)
	err ("e[uLpobt@mۏo܂ł.\n");
#else
    cur_expr = 0;		/* for Realloc() */
#endif

    while (1) {
	char	linebuf[1024];

	if( fgets( linebuf , sizeof linebuf , fp ) == NULL )
	    err ("Table  EOF ɒB܂.\n");

	if( *( strend( linebuf ) - 1 ) == '\n' )
	    *( strend( linebuf ) -1 ) = '\0';
	Line_num++;
	if( linebuf[0] == '#' || linebuf[0] == '*' )	/* comment */
	    continue;

	if( strnicmp( linebuf , "end" , 3 ) == 0 ) {
	    char	*ptr;
	    for( ptr = linebuf + 3 ; *ptr == ' ' || *ptr == '\t' ; ptr++ )
		;
	    if( *ptr == '[' ) {
		for( ptr++ ; *ptr == ' ' || *ptr == '\t' ; ptr++ )
		    ;
		if( isdigit( *ptr ) ) {
		    Table[ TableCounter ].loop = atoi( ptr );
		    break;
		} else if( strnicmp( ptr , "breakonly" , 9 ) == 0 ) {
		    /* e[u break ݂̂Ŕf */
		    Table[ TableCounter ].loop = TIMES_DECIDE_BY_BREAK;
		    break;
		} else if( *ptr == ']' ) {
		    Table[ TableCounter ].loop = TIMES_AUTOMATIC;
		    /* w肳Ȃ ->  */
		    break;
		}
		err ("Syntax error at end(%d s)\n", Line_num);
	    } else
		Table[ TableCounter ].loop = 1;
	    break;
	} else {	/* if( strnicmp( linebuf , "end" , 3 ) != 0 ) */

#ifdef	STATIC_TABLE_BUF
	    if (formula_line > MAX_TABLE_LINE)
		err ("TABLE ̍s܂.\n");

	    /* sobt@Ɋi[ */
	    cur_expr[formula_line].line = Line_num;
	    strcpy (cur_expr[formula_line].expr, linebuf);
#else
	    /* sobt@mۂĊi[ */
	    cur_expr = Realloc (cur_expr, sizeof (formula) * (formula_line + 1));
	    cur_expr[formula_line].line = Line_num;
	    cur_expr[formula_line].expr =
		strcpy (Malloc (strlen (linebuf) + 1), linebuf);
#endif
	    formula_line++;
	}
    }

#ifdef	STATIC_TABLE_BUF
    /* ]sobt@ */
    cur_expr = Realloc (cur_expr, sizeof (formula) * formula_line);
#endif
    Table[ TableCounter ].formulaptr = cur_expr;
    Table[ TableCounter ].lines      = formula_line;
    TableCounter++;
}


/*

  e[uLqt@Cǂݍ

*/
extern void	read_tablefile( char *filename )
{
    FILE* fp = fopen (filename, "rt");
    char  linebuf[256];

    if (!fp)
	err ("\n%s I[vł܂ł.\n", filename);

#ifdef	STATIC_TABLE_BUF
    Table = Malloc (sizeof (table) * MAX_TABLE_NUM);
#endif

    while( fgets( linebuf , sizeof linebuf , fp ) ) {
	Line_num++;
	if( linebuf[0] == '#' || linebuf[0] == '*' )	/* comment */
	    continue;
	if( isxdigit( linebuf[0] ) ) {
	    address	tabletop = (address)atox (linebuf);
	    eprintfq( "Table(%x)\n" , tabletop );
	    tablejob( tabletop , fp );
	}
    }
    interpret_all();
    fclose( fp );
}


/*

  e[ũT[`

*/
extern table	*search_table( address pc )
{
    int  i;

    for( i = 0 ; i < TableCounter ; i++ ) {
	if( Table[ i ].tabletop == pc )
	    return &Table[ i ];
    }
    return NULL;
}


/* EOF */
